#include <algorithm>
#include "parsetap.h"

ParseTap::ParseTap(string fileName,Log* outLog) : XMLTFAppParser(fileName,outLog)
{
   ivLog = outLog;

   LoadXmlSteps();  //inherited from XmlParser
   LoadTamFiles();
}

ParseTap::~ParseTap(void)
{
   
}

// Name: GetLogDisabled
// Purpose: To read from the teleflow xml Tap file the logdisable flag. The flag is
//          stored as numeric field number 1 or <n01>'flag'</n01>.
// Note: - Default value is zero. 
bool ParseTap::GetLogDisabled(int tamIndex) //fixed it
{
   int flag  = 0;
   int step = TamFilesStepID.at(tamIndex);
   
   GetFieldNumeric(step, 1, &flag);
   return (bool) flag;    
}

// Name: TapSize()
// Purpose: Returns the number of Tam Files found in the Tap file. 
int ParseTap::TapSize(void)
{
   return TamFiles.size();
}

// Name: TapParameterType
// Purpose: Retrieves the parameter type from the parameter field of the ProcStart
//          step in a teleflow tap file. The parameter type is specified by the
//          'PassedBy=' attribute.
// Note: - If passed by reference is found 1 is stored otherwise 0 is stored in '*num'
//       - 'int parameterValue' specifies which parameter field to retrieve the type from. 
int ParseTap::TapParameterType(int parameterValue, int* num)
{
   int functionNumber;                 //Will contain the function number of a teleflow step. 
   int unsigned teleFlowStep;          //index of the for loop.
   int unsigned steps = StepSize();    //obtains the total number of Steps in the Tap
   int parameterType;                  //The value read from the Parameter field of the Step
   TiXmlNode* pParent = NULL;          //The node pointer that will point to the step node found.

   for(teleFlowStep = 0; teleFlowStep < steps; teleFlowStep++)
   {
      if(GetRecordNumeric(teleFlowStep,"Function",&functionNumber) && functionNumber == 6)
      {
         pParent = VectorStepNodes.at(teleFlowStep);
         break;
      }
   }
   
  GetParamType(pParent,parameterValue,&parameterType);
  *num = parameterType;
  return 1;
}

// Name: TapParameterName
// Purpose: Retrieves the parameter name from the teleflow tap file. The Parameter name is
//          either a variable or constant. If no parameter name is found 'String& value' will be empty
//          otherwise it will contain the parameter name retrieved. 
int ParseTap::TapParameterName(int parameterValue,string&value)
{
   int functionNumber;
   int unsigned teleFlowStep;          //index of the for loop.
   int unsigned steps = StepSize();    //obtains the total number of Steps in the Tap
   string parameterString;             //The value read from the Parameter field of the Step
   TiXmlNode* pParent = NULL;          //The node pointer that will point to the step node found.

   for(teleFlowStep = 0; teleFlowStep < steps; teleFlowStep++)
   {
      if(GetRecordNumeric(teleFlowStep,"Function",&functionNumber) && functionNumber == 6)
      {
         pParent = VectorStepNodes.at(teleFlowStep);
         break;
      }
   }
   
  GetParamName(pParent,parameterValue,parameterString);

  if(parameterString == "")
  {
      return 0;
  }
  else
  {
      value = parameterString; 
      return 1;
  }
}

// Name: TapParameterValue
// Purpose: Checks if the number specified by 'int parameterValue' is a valid parameter field in
//          the tap file. If the parameterfield with the specific parameterValue is found true is
//          returned otherwise false is returned. 
// Note: - Parameters can only appear in the ProcStart step of tap file.
//       - 'in parameterValue' should be between 1 and 9
bool ParseTap::TapParameterValue(int parameterValue, string& value)//the basic search function for the parameter.
{
   int functionNumber;                 //The function number of the step.
   int unsigned teleFlowStep;          //index of the for loop.
   int unsigned steps = StepSize();    //obtains the total number of Steps in the Tap
   string parameterString;             //The value read from the Parameter field of the Step
   TiXmlNode* pParent = NULL;          //The node pointer that will point to the step node found.

   for(teleFlowStep = 0; teleFlowStep < steps; teleFlowStep++)
   {
      if(GetRecordNumeric(teleFlowStep,"Function",&functionNumber) && functionNumber == 6)
      {
         pParent = VectorStepNodes.at(teleFlowStep);
         break;
      }
   }

  GetParamValue(pParent,parameterValue,parameterString);

  if(parameterString == "")
  {
      return 0;
  }
  else
  {
      value = parameterString; 
      return 1;
  }
} 

// Name: FixFilePath
// Purpose: To potentially fix the filepath of the tam loaded from the teleflow xml.
//          - If the Tam filepath has a '..\' it will change the tap file path so that
//             the the Tam file can be found. 
void ParseTap::FixFilePath(string& file)
{
   string tapPath = XMLTFAppParser::filePath;//FilePath to .Tap
   size_t pos;
   int end;

   pos = tapPath.find_last_of("\\"); 

   if( pos != string::npos)//get the directory
   {
      end = tapPath.length();
      tapPath.erase(pos,end);
   }
   else
   {
      tapPath.clear(); //can only contain the filenam not the path, so it can be cleared.
   }
   
   while(file.find("..\\") != string::npos)
   {
      pos = tapPath.find_last_of("\\",end-1);

      if(pos != string::npos)
      {
         end = tapPath.length();
         tapPath.erase(pos,end);
      }

      file.erase(0,3);
   }

   if(!tapPath.empty())
   {
      tapPath += "\\";
   }

   tapPath += file;
   file = tapPath;
}

// Name: LoadTamFiles
// Purpose: Loads every Tam file path and step number into a vector. 
void ParseTap::LoadTamFiles(void)
{
   int unsigned stepSize = StepSize(); //Number of steps in the TAP File.
   string elementData;
   TiXmlNode* pChild;
   TiXmlElement* pElement;
   int number;

   for(int unsigned index =0;index < stepSize;index++) //looping over steps.,
   {     
      if(GetRecordNumeric(index,"Function",&number) && number == 3)//better way to search.
      {
         if(GetFieldString(index, 1, elementData) == 1)
         {
            ConvertToXmlExt(elementData);//fixed the .tam or .Tam or .TAM bug.
            TamFiles.push_back(elementData);
            TamFilesStepID.push_back(index);
            elementData.clear();
         }
      }

      pChild = VectorStepNodes.at(index);
      pElement = pChild->ToElement();

      if(pElement->QueryValueAttribute("Function",&number) != TIXML_NO_ATTRIBUTE && number == 2)
      {
         if(GetFieldString(index,1,elementData) == 1)
         {   
            voiceDbf = elementData;
         }
         if(GetFieldString(index, 2,elementData) == 1)
         {
            searchVoiceDbf = elementData;
         }
      }
   }
}

// Name: TapGetOpCode
// Purpose: Retrieves the opCode value from the teleflow step in a tap xml file.
//          The OpCode is the 'function=' attribute in xml.
// Note: - 'int tamIndex' specifies the Node to a teleflow step with a Tam filepath. 
int ParseTap::TapGetOpCode(int tamIndex, string& data)
{
   //Opcode is Function in XML
   int tamStep = 0;

   tamStep = TamFilesStepID.at(tamIndex);
   if(GetRecordString(tamStep,"Function",data) == 0)
   {
      return 0;
   }
   
   return 1; 
}

// Name: TapGetLabel
// Purpose: Retrieves the GetLabel value from the teleflow step in a tap xml file.
//          The Getlabel is the 'Label=' attribute in the xml file. 
int ParseTap::TapGetLabel(int tamIndex, string& data)
{
   int tamStep = 0;

   tamStep = TamFilesStepID.at(tamIndex);

   if(GetRecordString(tamStep,"Label",data) == 0)
   {
      return 0;
   }
   
   return 1;
}

// Name: TapGetStepNum
// Purpose: Retrieves the GetStepNum value from the teleflow step in a tap xml file.
//          The GetStepNum is the 'Current=' attribute in the xml file. 
int ParseTap::TapGetStepNum(int tamIndex, string& data)
{
   int tamStep = 0;

   tamStep = TamFilesStepID.at(tamIndex);

   if(GetRecordString(tamStep, "Current",data) == 0)
   {
      return 0;
   }
   
   return 1; 
}

// Name: FixFilename
// Purpose: To change the filename in the xml DBF file to the actual file stored on disk. 
int ParseTap::FixVoiceFileName(string& file)
{
   size_t position;
   
   std::transform(file.begin(),file.end(),file.begin(),toupper);
   position = file.find(".DBF");

   if(position == std::string::npos)
   {  
      return 0;
   }
   else
   {
      file.erase(position, 4);
      file += "_dbf.xml";
      return 1;
   }
}

// Name: ConvertToXmlExt
// Purpose: To convert the file extension of the tam file in the teleflow file from
//          '.Tam' to '.xml'.
void ParseTap::ConvertToXmlExt(string& filePath)
{
   size_t position;
   std::transform(filePath.begin(),filePath.end(),filePath.begin(),toupper);

   position = filePath.find(".TAM");

   filePath.erase(position, 4);
   filePath += ".xml";
}

// Name: TamAt
// Purpose: Returns the Tam filepath at the index specified by 'int tamIndex'. 
string ParseTap::TamAt(int tamIndex)
{
   return TamFiles.at(tamIndex);//returns the Tamfilename
}

//retrieves the voice.dbf file
string ParseTap::GetVoiceFile()
{
   return voiceDbf;
}
//retrieves the search path for the voice.dbf file. 
string ParseTap::GetSearchVoiceFile(void)
{
   return searchVoiceDbf;
}

        



