#ifndef STEP_RUNBASICSCRIPT_H
#define STEP_RUNBASICSCRIPT_H

#include <string>
using std::string;
#include <map>
using std::map;
#include <list>
using std::list;
#include <set>
using std::set;
#include <vector>
using std::vector;

#include "step.h"

// Class to handle script for parsing
class Script
{
   public:

      // Each token must be one of the following types
      // If the script has reached the end, the token type will be NONE
      // Any token that is not recognised as one of the other types will be of type INVALID
      enum TOKEN_TYPE
      {
         NONE,
         INVALID,
         OPERATOR,
         SEPARATOR,
         NUMBER,
         NAME,
         FUNCTION
      };

      Script(string newScript);
      virtual ~Script(void);

      // Returns the current token if there is one
      // If there is no token, an empty string is returned and the type will be NONE
      // Tokens are returned in upper case
      string GetCurrentToken(void);

      // Returns the type of the current token
      Script::TOKEN_TYPE GetCurrentTokenType(void);

      // Resets the script to the beginning and returns the first token as per GetCurrentToken()
      string GetFirstToken(void);

      // Moves forward in the script to the next token and returns that token as per GetCurrentToken()
      string GetNextToken(void);

      // Returns the previous expression token if there was one
      // If there was no token, an empty string is returned and the type will be NONE
      // Tokens are returned in upper case
      string GetPreviousExpressionToken(void);

      // Returns the type of the previous expression token
      Script::TOKEN_TYPE GetPreviousExpressionTokenType(void);

      // Returns a string of all text until the next quote or end of script
      // Retains case
      string GetUntilEndQuote(void);

      // Resets previous expression token to an empty string and the previous expression token type to NONE
      // Moves forward in the script to the next token and returns that token as per GetCurrentToken()
      string StartNewExpression(void);

   protected:

      string script;
      string::iterator parseIndex;
      string currentToken;
      TOKEN_TYPE currentTokenType;
      string previousExpressionToken;
      TOKEN_TYPE previousExpressionTokenType;
};

// Pure virtual base class to construct a parse tree for expressions
// All branches are operations or parentheses, all leaves are constants or variables
// Expressions have left and right branches which are of higher or equal mathematical precendence
// In the case of equal precedence and filled branches, the expression is added to the right branch
// NOTE: The destructor, Add, Evaluate, and Dump are all recursive methods
class Expression
{
   public:

      Expression(int priority);
      virtual ~Expression(void);

      // Add an expression to another expression or expression tree and follow
      // rules of mathematical precedence for construction tree
      // Returns the new tree
      virtual Expression* Add(Expression* expression);

      // Returns the value that an expression tree evaluates to by preorder traversal
      // If invalid is false when passed in, it will only be true afterwards if
      // an error has occurred
      // May return an empty string if there is an error
      virtual string Evaluate(PIVRINFO pInfo, bool& invalid) = 0;

      // Debugging function that returns the name of an expression
      virtual string Name(void) = 0;

      // Debugging function that displays the addresses of the expressions in the
      // expression tree in a preorder traversal
      void Dump(Log* ivLog);

      // Returns the mathematical precendence of the expression, from lowest (1) to highest (13)
      // 1:  OR
      // 2:  XOR
      // 3:  AND
      // 4:  BOR
      // 5:  BXOR
      // 6:  BAND
      // 7:  =, !=, <>
      // 8:  <, <=, >, >=
      // 9:  Additive operators: +, -
      // 10: *, /, MOD
      // 11: ^
      // 12: Unary operators: +, -, NOT, BNOT
      // 13: Constant, Variable, Function
      // 14: Parentheses
      int GetPriority(void);

      virtual string GetString(void) const;

      virtual string Symbol(void) const;

   protected:

      // Takes the evaluated left and right trees and uses TeleFlow's compare
      // function to generate the return string
      // If invalid is false when passed in, it will only be true afterwards if
      // an error has occurred
      // May return an empty string if there is an error
      string EvaluateCompare(PIVRINFO pInfo, CMP_FN compareFunction, bool& invalid);

      // Takes the evaluated left and right trees and uses TeleFlow's math
      // function to generate the return string
      // If invalid is false when passed in, it will only be true afterwards if
      // an error has occurred
      // May return an empty string if there is an error
      string EvaluateMath(PIVRINFO pInfo, MATH_FN mathFunction, bool& invalid);

      Expression* left;
      Expression* right;
      int priority;
};

// Pure virtual base class to represent functions
// Functions have a predetermined number of parameter Expressions
// All Functions have the same priority
class Function : public Expression
{
   public:

      // Constructor: specify how many expected parameters in paramsCount
      Function(unsigned int paramsCount);
      virtual ~Function(void);

      // Adds a single Expression* to the params vector
      void AddParameter(Expression* expression);

      // Returns the amount of expected parameters
      unsigned int GetExpectedParamsCount(void);

      virtual string GetString(void) const;

   protected:

      // Populates evaluatedParams in function of params at runtime
      bool EvaluateParameters(PIVRINFO pInfo, bool& invalid);

      unsigned int expectedParamsCount;
      vector<string> evaluatedParams;
      vector<Expression*> params;
};

// Pure virtual base class to represent unary operators
// Unary operators only operate on the Expression to the right
// All unary operators have the same priority
class UnaryOperator : public Expression
{
   public:

      UnaryOperator(void);
      virtual ~UnaryOperator(void);

      // Unary operators only add to the right
      Expression* Add(Expression* expression);

      virtual string GetString(void) const;
};

// Utility class to make sure that memory allocated in trees on the stack is properly
// de-allocated when the stack destructs
class ExpressionTreeStack : public deque<Expression*>
{
   public:

   ExpressionTreeStack(void);
   virtual ~ExpressionTreeStack(void);
};

class AbsoluteValue : public Function
{
   public:

      AbsoluteValue(void);
      virtual ~AbsoluteValue(void);

      // Returns the absolute value of <parameter>
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Add : public Expression
{
   public:

      Add(void);
      virtual ~Add(void);

      // Returns the sum of the left and right expression trees
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class BitwiseAnd : public Expression
{
   public:

      BitwiseAnd(void);
      virtual ~BitwiseAnd(void);

      // Returns the bitwise AND of the left and right expression trees
      // The values are converted to integers for the comparison and non-numeric
      // strings will be interpreted as 0
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;

   protected:

      string value;
};

class LogicalAnd : public Expression
{
   public:

      LogicalAnd(void);
      virtual ~LogicalAnd(void);

      // Returns the logical AND of the left and right expression trees
      // The values are converted to integers for the comparison and non-numeric
      // strings will be interpreted as 0
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;

   protected:

      string value;
};

class Asc : public Function
{
   public:

      Asc(void);
      virtual ~Asc(void);

      // Returns the character that corresponds to the ASCII value <parameter>
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Char : public Function
{
   public:

      Char(void);
      virtual ~Char(void);

      // Returns the character that corresponds to the ASCII value <parameter>
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Char2Hex : public Function
{
   public:

      Char2Hex(void);
      virtual ~Char2Hex(void);

      // Returns the character that corresponds to the ASCII value <parameter>
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Concat : public Function
{
   public:

      Concat(void);
      virtual ~Concat(void);

      // Returns <first parameter> concatenated with <second parameter>
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Constant : public Expression
{
   public:

      Constant(string value);
      virtual ~Constant(void);

      // Returns a parsed version of the value that Constant is constructed with
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string GetString(void) const;

      string Name(void);

   protected:

      string value;
};

class Divide : public Expression
{
   public:

      Divide(void);
      virtual ~Divide(void);

      // Returns the quotient of the left expression tree with the right
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class EqualTo : public Expression
{
   public:

      EqualTo(void);
      virtual ~EqualTo(void);

      // Returns "1" if the left and right expression trees are equal, and "0" if they are not
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Exponentiate : public Expression
{
   public:

      Exponentiate(void);
      virtual ~Exponentiate(void);

      // Returns the left expression to the right expression power
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class GreaterThan : public Expression
{
   public:

      GreaterThan(void);
      virtual ~GreaterThan(void);

      // Returns "1" if the left expression tree is greater than the right, and "0" if it is not
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class GreaterThanOrEqualTo : public Expression
{
   public:

      GreaterThanOrEqualTo(void);
      virtual ~GreaterThanOrEqualTo(void);

      // Returns "1" if the left expression tree is greater than or equal to the right, and "0" if it is not
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Hex2Char : public Function
{
   public:

      Hex2Char(void);
      virtual ~Hex2Char(void);

      // Returns the character that corresponds to the ASCII value <parameter>
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Left : public Function
{
   public:

      Left(void);
      virtual ~Left(void);

      // Returns the left <second parameter> characters of <first parameter>
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Len : public Function
{
   public:

      Len(void);
      virtual ~Len(void);

      // Returns the length of <parameter>
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class LessThan : public Expression
{
   public:

      LessThan(void);
      virtual ~LessThan(void);

      // Returns "1" if the left expression tree is less than the right, and "0" if it is not
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class LessThanOrEqualTo : public Expression
{
   public:

      LessThanOrEqualTo(void);
      virtual ~LessThanOrEqualTo(void);

      // Returns "1" if the left expression tree is less than or equal to the right, and "0" if it is not
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Lowercase : public Function
{
   public:

      Lowercase(void);
      virtual ~Lowercase(void);

      // Returns <parameter> with all characters converted to lower case
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Mod : public Expression
{
   public:

      Mod(void);
      virtual ~Mod(void);

      // Returns the modulus of the left expression tree with the right
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Multiply : public Expression
{
   public:

      Multiply(void);
      virtual ~Multiply(void);

      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class LogicalNot : public UnaryOperator
{
   public:

      LogicalNot(void);
      virtual ~LogicalNot(void);

      // Returns "1" if the right expression tree evaluates to "0", and "0" if it doesn't
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class BitwiseNot : public UnaryOperator
{
   public:

      BitwiseNot(void);
      virtual ~BitwiseNot(void);

      // Returns the ones complement of the right expression tree
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Negate : public UnaryOperator
{
   public:

      Negate(void);
      virtual ~Negate(void);

      // Returns the negative value of the right expression tree
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class NotEqualTo : public Expression
{
   public:

      NotEqualTo(void);
      virtual ~NotEqualTo(void);

      // Returns "1" if the left and right expression trees are not equal, and "0" if they are
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class BitwiseExclusiveOr : public Expression
{
   public:

      BitwiseExclusiveOr(void);
      virtual ~BitwiseExclusiveOr(void);

      // Returns the bitwise exclusive OR of the left and right expression trees
      // The values are converted to integers for the comparison and non-numeric
      // strings will be interpreted as 0
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class LogicalExclusiveOr : public Expression
{
   public:

      LogicalExclusiveOr(void);
      virtual ~LogicalExclusiveOr(void);

      // Returns the logical OR of the left and right expression trees
      // The values are converted to integers for the comparison and non-numeric
      // strings will be interpreted as 0
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class BitwiseOr : public Expression
{
   public:

      BitwiseOr(void);
      virtual ~BitwiseOr(void);

      // Returns the bitwise OR of the left and right expression trees
      // The values are converted to integers for the comparison and non-numeric
      // strings will be interpreted as 0
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class LogicalOr : public Expression
{
   public:

      LogicalOr(void);
      virtual ~LogicalOr(void);

      // Returns the logical OR of the left and right expression trees
      // The values are converted to integers for the comparison and non-numeric
      // strings will be interpreted as 0
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

// Used to provide mathematical precedence protection for parenthesised expressions
class Parenthesised : public Expression
{
   public:

      Parenthesised(Expression* expression);
      virtual ~Parenthesised(void);

      // Returns the value of the left expression tree that Parenthesised is constructed with
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string GetString(void) const;

      string Name(void);
};

class Pos : public Function
{
   public:

      Pos(void);
      virtual ~Pos(void);

      // Returns the position of the first <second parameter> in <first parameter>
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Right : public Function
{
   public:

      Right(void);
      virtual ~Right(void);

      // Returns the right <second parameter> characters of <first parameter>
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Round : public Function
{
   public:

      Round(void);
      virtual ~Round(void);

      // Returns <parameter> rounded to the nearest integer
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Sign : public Function
{
   public:

      Sign(void);
      virtual ~Sign(void);

      // Returns 1 if <parameter> is positive
      // Returns 0 if <parameter> is zero
      // Returns -1 if <parameter> is negative
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class SubStr : public Function
{
   public:

      SubStr(void);
      virtual ~SubStr(void);

      // Returns the <third parameter> characters of <first parameter> starting at character <second parameter>
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Subtract : public Expression
{
   public:

      Subtract(void);
      virtual ~Subtract(void);

      // Returns the difference of the left and right expression trees
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Trim : public Function
{
   public:

      Trim(void);
      virtual ~Trim(void);

      // Returns <parameter> with the spaces trimmed from the front and back
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Uppercase : public Function
{
   public:

      Uppercase(void);
      virtual ~Uppercase(void);

      // Returns <parameter> with all characters converted to upper case
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class DatePart : public Function
{
   public:

      DatePart(void);
      virtual ~DatePart(void);

      // Returns the date/time equivalent of a given Unix time <first parameter>
      // adjusted by the UTC offset <second parameter> in seconds
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class TimePart : public Function
{
   public:

      TimePart(void);
      virtual ~TimePart(void);

      // Returns the date/time equivalent of a given Unix time <first parameter>
      // adjusted by the UTC offset <second parameter> in seconds
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string Name(void);
      string Symbol(void) const;
};

class Variable : public Expression
{
   public:

      Variable(string variableName);
      virtual ~Variable(void);

      // Returns the current value of the variable who's name Variable is constructed with
      string Evaluate(PIVRINFO pInfo, bool& invalid);

      string GetString(void) const;

      string Name(void);

   protected:

      string variableName;
};

// Pure virtual base class for each programmatic command in the linked list that
// represents the BASIC script
class Command
{
   public:

      Command(void);
      virtual ~Command(void);

      // Debugging function that displays the addresses of the commands in the
      // command list and any commands that are linked to it
      virtual void Dump(Log* ivLog);

      // Returns true if no errors occur, returns false otherwise
      virtual bool Execute(PIVRINFO pInfo) = 0;

      // Gets the link to the next command in the linked list of programmatic
      // commands that represents the BASIC script
      virtual Command* GetNextCommand(void);

      // Sets the link to the next command in the linked list of programmatic
      // commands that represents the BASIC script
      virtual void SetNextCommand(Command* nextCommand);

      // Debugging function that returns the name of a command
      virtual string Name(void) = 0;

   protected:

      Command* next;
};

class Branch : public Command
{
   public:

      typedef enum
      {
         CASE,
         DO,
         FOR,
         IF,
         WHILE
      } BRANCH_TYPE;

      Branch(BRANCH_TYPE dType);
      virtual ~Branch(void);

      BRANCH_TYPE GetBranchType(void);
      Command* GetElseBranch(void);
      Command* GetIfBranch(void);
      void SetElseBranch(Command* command);
      void SetNextCommand(Command* command);

   protected:

      Command* ifBranch;
      Command* elseBranch;
      BRANCH_TYPE branchType;
};

class Assignment : public Command
{
   public:

      Assignment(string variableName, Expression* expressionTree);
      virtual ~Assignment(void);

      // Assigns the variable variableName to the evaluation of expressionTree
      virtual bool Execute(PIVRINFO pInfo);

      string Name(void);

   protected:

      string variableName;
      Expression* expressionTree;
};

class ScopedAssignment : public Assignment
{
   public:

      ScopedAssignment(string variableName, Expression* expressionTree, bool global);
      virtual ~ScopedAssignment(void);

      // Assigns the variable variableName to the evaluation of expressionTree
      bool Execute(PIVRINFO pInfo);

   protected:

      bool global;
};

class Break : public Command
{
   public:

      Break(void);
      virtual ~Break(void);

      bool Execute(PIVRINFO pInfo);

      string Name(void);
};

class Case : public Branch
{
   public:

      Case(Expression* argument, Expression* lowerBoundCondition, Case* linkedCase = NULL);
      virtual ~Case(void);

      bool Execute(PIVRINFO pInfo);
      Expression* GetArgument(void);
      string Name(void);

      // We need to take care of all linked case commands
      virtual void SetNextCommand(Command* nextCommand);

      void SetUpperBoundCondition(Expression* condition);

   protected:

      Case *previousCaseLinked;
      Expression* argument;
      Expression* lowerBoundCondition;
      Expression* upperBoundCondition;
};

class CaseElse : public Case
{
   public:

      CaseElse(void);
      virtual ~CaseElse(void);

      string Name(void);
};

class Do : public Branch
{
   public:

      Do(void);
      virtual ~Do(void);

      // Decides which command to execute next based on the evaluation of the
      // expression that Branch is constructed with
      bool Execute(PIVRINFO pInfo);

      string Name(void);
};

class Else : public Branch
{
   public:

      Else(void);
      virtual ~Else(void);

      // Decides which command to execute next based on the evaluation of the
      // expression that Branch is constructed with
      bool Execute(PIVRINFO pInfo);

      string Name(void);
};

class End : public Command
{
   public:

      End(void);
      virtual ~End(void);

      // Decides which command to execute next based on the evaluation of the
      // expression that Branch is constructed with
      bool Execute(PIVRINFO pInfo);

      virtual string Name(void);
};

class For : public Branch
{
   public:

      For(string variableName, Expression* start, Expression* end, Expression* step);
      virtual ~For(void);

      // Evaluates the loopStart, loopEnd, and step expression trees, passes that
      // information to the NEXT that closes the loop, and then proceeds into the loop
      bool Execute(PIVRINFO pInfo);

      string Name(void);

   protected:

      string variableName;
      Expression* start;
      Expression* end;
      Expression* step;
};

class GoSub : public Command
{
   public:

      GoSub(deque<Command*>& returnStack);
      virtual ~GoSub(void);

      void Dump(Log* ivLog);

      // Places returnNext on the returnStack, and jumps to the Label called by GoSub
      bool Execute(PIVRINFO pInfo);

      string Name(void);

      // This method must be called twice: first when the Command that follows
      // the GoSub is parsed, and second when the Label called by GoSub is parsed
      // The first call sets next to the Command that follows the GoSub and returnNext
      // to NULL
      // The second call sets next to the Label called by GoSub, and sets returnNext
      // to the Command that follows GoSub
      void SetNextCommand(Command* nextCommand);

   private:

      Command* returnNext;
      deque<Command*>& returnStack;
};

class GoTo : public Command
{
   public:

      GoTo(void);
      virtual ~GoTo(void);

      // Goes to the Label called by GoTo
      bool Execute(PIVRINFO pInfo);

      string Name(void);
};

class If : public Branch
{
   public:

      If(Expression* condition);
      virtual ~If(void);

      virtual void Dump(Log* ivLog);

      // Decides which command to execute next based on the evaluation of the
      // expression that Branch is constructed with
      virtual bool Execute(PIVRINFO pInfo);

      virtual string Name(void);

      Expression* GetCondition(void);
      void SetCondition(Expression* condition);

   protected:

      Expression* condition;
};

class ElseIf : public If
{
   public:

      ElseIf(Expression* condition);
      virtual ~ElseIf(void);

      string Name(void);
};

class Label : public Command
{
   public:

      Label(void);
      virtual ~Label(void);

      // No-op Command for FOR-NEXT, GOSUB, and GOTO
      bool Execute(PIVRINFO pInfo);

      string Name(void);
};

class Next : public Branch
{
   public:

      Next(void);
      virtual ~Next(void);

      // Evaluates the loopStart, loopEnd, and step expression trees, passes that
      // information to the NEXT that closes the loop, and then proceeds into the loop
      bool Execute(PIVRINFO pInfo);

      void SetEndValue(int value);
      void SetStartValue(int value);
      void SetStepValue(int value);
      void SetVariableName(string name);

      string Name(void);

   protected:

      string variableName;
      int startValue;
      int endValue;
      int stepValue;
};

class Print : public Command
{
   public:

      Print(Constant* printString);
      virtual ~Print(void);

      // Prints the value of printString to the log
      bool Execute(PIVRINFO pInfo);

      string Name(void);

   private:

      Constant* printString;
};

class Return : public Command
{
   public:

      Return(deque<Command*>& returnStack);
      virtual ~Return(void);

      // Pops the top Command off of returnStack and executes it
      bool Execute(PIVRINFO pInfo);

      string Name(void);

   private:

      deque<Command*>& returnStack;
};

class Select : public Case
{
   public:

      Select(Expression* argument);
      virtual ~Select(void);
      
      string Name(void);
};

class While : public If
{
   public:

      While(Expression* condition);
      virtual ~While(void);

      string Name(void);
};

class Until : public While
{
   public:

      Until(Expression* condition);
      virtual ~Until(void);

      string Name(void);
};

class StepRunBASICScript : public Step
{
   public:

      StepRunBASICScript(STEPREC& stepRecord);
      virtual ~StepRunBASICScript(void);

		const char* GetStepName(void) const;

      // Parses script on the first execution
      // Executes linked list of commands that represent BASIC script
      IV_STATUS Execute(PIVRINFO pInfo);

      // Parses and expression until a non-expression token and returns an
      // expression tree that follows mathematical precedence
      Expression* ParseExpression(Log* ivLog, bool defensiveParenthesisReturn = false);

      // Parses the script to construct the linked list of Commands
      // Commands are placed in the commands list so that they can be deallocated
      // when StepRunBASICScript destructs
      // IMPORTANT NOTE:  Although the first Command in commands is the first Command
      // in the linked list, the ordering of commands has no other relevance other
      // than the order in which the Commands were created, and cannot be relied
      // upon to provide order of execution
      bool ParseCommands(Log* ivLog);

   protected:

      Script BASICScript;

      // List of commands for clean up purposes and to hold the first Command in
      // the linked list that represents the BASIC script
      list<Command*> commands;

      // List of keywords that cannot be used as labels
      set<string> keywords;

      // Stack of Commands for returns from GoSubs
      deque<Command*> returnStack;

      // Flag to indicate that the script has been parsed and does not need to be
      // parsed a second time
      bool parsedScript;
};

#endif
