#include <stdlib.h>             // Needed for "atoi".

#include <string>
using std::string;

#include "step_placecall.h"

StepPlaceCall::StepPlaceCall(STEPREC& stepRecord) : Step(stepRecord)
{
   ring = stepRecord.s3.arr;
   phone = stepRecord.s2.arr;
   callerID = stepRecord.s4.arr;
   flashDuration = stepRecord.s5.arr;
   voiceLengthVariable = stepRecord.s6.arr;
   silenceTimeout = stepRecord.s7.arr;
   maxVoiceTimeout = stepRecord.s8.arr;
   bSameLine = (bool) stepRecord.u1;
   //
   // ISDN only calling plan/type fields
   isdnCallingNumberPlan = stepRecord.s9.arr;
   isdnCallingNumberType = stepRecord.s10.arr;
   // ISDN called plan/type fields combined in S1 due to step field constraints
   isdnCalledNumParms = stepRecord.s1.arr;
   //
   // Default value of 0 indicates connect on CED (i.e. true)
   bConnectOnCED = !((bool) stepRecord.u2);
   bConnectOnSIT = (bool) stepRecord.u3;
}

StepPlaceCall::~StepPlaceCall(void)
{
}

const char* StepPlaceCall::GetStepName(void) const
{
	return("Place Call");
}

// -------------------------------------------------------------------------
//   Name:        StepPlaceCall::Execute
//
//   Modified:    Jul 08/02 - TF: Changed to use global @PORT setting from
//                                version 4 TAL file
//                Dec 10/02 - TF: Added code to retrieve Caller ID to send
//                                with the call
//                Mar 21/03 - TF: Flash hook for "use current connection"
//                                can be set by application designer
// -------------------------------------------------------------------------
IV_STATUS StepPlaceCall::Execute(PIVRINFO pInfo)
{
   int iRingCnt;
   string parsedRingCount;
   IV_STATUS status;
   char* value;
   // Evaluate all parameters before exiting
   bool bParmsValid = true;
   string message;
   string localPhone;
   char* localCallerID = NULL;
   int   iFlashDur = 0;
   string parsedFlashDuration;
   int voiceLength;
   int* pVoiceLength = NULL;
   int iSilenceTimeout = 0;
   int iMaxVoiceTimeout = 0;
   BoardGeneric::ISDNNumberPlan iCallingNumPlan = BoardGeneric::UNKNOWN_NUMBER_PLAN;
   BoardGeneric::ISDNNumberType iCallingNumType = BoardGeneric::UNKNOWN_NUMBER_TYPE;
   //
   // ISDN called plan/type fields, must be parsed from 'isdnCalledNumParms';
   // (had to be combined due to step field constraints)
   string isdnCalledNumberPlan = "";
   string isdnCalledNumberType = "";
   size_t fieldStart;
   size_t fieldEnd;
   BoardGeneric::ISDNNumberPlan iCalledNumPlan = BoardGeneric::UNKNOWN_NUMBER_PLAN;
   BoardGeneric::ISDNNumberType iCalledNumType = BoardGeneric::UNKNOWN_NUMBER_TYPE;

   if(Status(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FAIL);
   }

	if(PortRequired(pInfo) != IVR_SUCCESS)
   {
      return(IVR_FATAL);
   }

   if((value = pInfo->variables->ParseParameter("Telephone Number", phone.c_str())) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      localPhone = value;
      delete[] value;
   }

   if((value = pInfo->variables->ParseParameter("Rings before TIMEOUT", ring.c_str(), "4")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      parsedRingCount = value;
      iRingCnt = atoi(value);
      delete[] value;

      if(iRingCnt <= 0)
      {
         pInfo->ivLog->Print("      ERROR: Rings before TIMEOUT '%s' must be at least 1\n", ring.c_str());
         bParmsValid = false;
      }
   }

   if(!callerID.empty() && ((localCallerID = pInfo->variables->ParseParameter("Caller ID", callerID.c_str(), "")) == NULL))
   {
      bParmsValid = false;
   }

   if(bSameLine)
   {
      if((value = pInfo->variables->ParseParameter("On-hook time", flashDuration.c_str(), "500")) == NULL)
      {
         bParmsValid = false;
      }
      else
      {
         parsedFlashDuration = value;
         iFlashDur = atoi(value);
         delete[] value;
      }

      if(iFlashDur < 0)
      {
         pInfo->ivLog->Print("      ERROR: On-hook time '%s' must be >= 0\n", flashDuration.c_str());
         return(IVR_FATAL);
      }
   }

   if(!voiceLengthVariable.empty())
   {
      if(!pInfo->variables->IsVariableFormat(voiceLengthVariable.c_str()))
      {
         pInfo->ivLog->Print("      ERROR: Invalid voice length variable '%s'\n", voiceLengthVariable.c_str());
         bParmsValid = false;
      }
      else if(pInfo->variables->IsSystem(voiceLengthVariable.c_str()))
      {
         pInfo->ivLog->Print("      ERROR: Cannot use system variable '%s' for voice length variable\n", voiceLengthVariable.c_str());
         bParmsValid = false;
      }
      pInfo->ivLog->Print("      Capture answering voice length to %s\n", voiceLengthVariable.c_str());
      pVoiceLength = &voiceLength;
   }
   else
   {
      pInfo->ivLog->Print("      Do not capture answering voice length to TeleFlow variable\n");
      // pVoiceLength stays NULL
   }

   if((value = pInfo->variables->ParseParameter("Silence timeout", silenceTimeout.c_str(),
      "0", "Empty Silence timeout defaults to none")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iSilenceTimeout = atoi(value);
      delete[] value;
      // zero is valid and means no voice detection
   }

   if((value = pInfo->variables->ParseParameter("Maximum voice length timeout", maxVoiceTimeout.c_str(),
      "0", "Empty Maximum voice length timeout defaults to none")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      iMaxVoiceTimeout = atoi(value);
      delete[] value;
      // zero is valid and means no voice detection
   }

   if(((iSilenceTimeout == 0) && (iMaxVoiceTimeout != 0)) || ((iSilenceTimeout != 0) && (iMaxVoiceTimeout == 0)))
   {
      pInfo->ivLog->Print("      ERROR: Silence Timeout and Maximum Voice Length must both be 0 or empty\n");
      pInfo->ivLog->Print("             to disable voice detection\n");
      bParmsValid = false;
   }

   // Evaluate ISDN Calling Number Plan parameter
   if((value = pInfo->variables->ParseParameter("ISDN Calling Number Plan", isdnCallingNumberPlan.c_str(),
      "DEFAULT", "Empty Calling Number Plan set to DEFAULT")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      if(EvaluateNumberPlan(value, iCallingNumPlan) == false)
      {
         pInfo->ivLog->Print("      ERROR: Invalid ISDN calling plan '%s'\n", value);
         bParmsValid = false;
      }
      delete[] value;
   }

   // Evaluate ISDN Calling Number Type parameter
   if((value = pInfo->variables->ParseParameter("ISDN Calling Number Type", isdnCallingNumberType.c_str(),
      "DEFAULT", "Empty Calling Number Type set to DEFAULT")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      if(EvaluateNumberType(value, iCallingNumType) == false)
      {
         // invalid
         pInfo->ivLog->Print("      ERROR: Invalid ISDN calling type '%s'\n", value);
         bParmsValid = false;
      }
      delete[] value;
   }

   // Parse ISDN called number parms into separate fields, isdnCalledNumberPlan & isdnCalledNumberType
   // (END_SUBFIELD is defined in ivr_defs.h)
   fieldStart = 0;
   fieldEnd = isdnCalledNumParms.find(END_SUBFIELD, fieldStart);
   if(fieldEnd != string::npos)
   {
      isdnCalledNumberPlan = isdnCalledNumParms.substr(fieldStart, fieldEnd - fieldStart);
      fieldStart = fieldEnd + 1;
      fieldEnd = isdnCalledNumParms.find(END_SUBFIELD, fieldStart);
      if(fieldEnd != string::npos)
      {
         isdnCalledNumberType = isdnCalledNumParms.substr(fieldStart, fieldEnd - fieldStart);
      }
   }

   // Evaluate ISDN Called Number Plan parameter
   if((value = pInfo->variables->ParseParameter("ISDN Called Number Plan", isdnCalledNumberPlan.c_str(),
      "DEFAULT", "Empty Called Number Plan set to DEFAULT")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      if(EvaluateNumberPlan(value, iCalledNumPlan) == false)
      {
         pInfo->ivLog->Print("      ERROR: Invalid ISDN called plan '%s'\n", value);
         bParmsValid = false;
      }
      delete[] value;
   }

   // Evaluate ISDN Called Number Type parameter
   if((value = pInfo->variables->ParseParameter("ISDN Called Number Type", isdnCalledNumberType.c_str(),
      "DEFAULT", "Empty Called Number Type set to DEFAULT")) == NULL)
   {
      bParmsValid = false;
   }
   else
   {
      if(EvaluateNumberType(value, iCalledNumType) == false)
      {
         pInfo->ivLog->Print("      ERROR: Invalid ISDN called type '%s'\n", value);
         bParmsValid = false;
      }
      delete[] value;
   }

   if(!bParmsValid)
   {
      delete[] localCallerID;
      return(IVR_FATAL);
   }

   message = (string) "      Place call to " + localPhone;
   if(bSameLine)
   {
      message += (string) "\n      on the current connection using a flash hook length of " + parsedFlashDuration;
   }
   message += (string) "\n      and wait for up to " + parsedRingCount + (string) " rings";

   pInfo->ivLog->Print("%s\n", message.c_str());

   pInfo->ivLog->Print("      %sonnect on CED\n", bConnectOnCED ? "C" : "Do not c");
   pInfo->ivLog->Print("      %sonnect on SIT\n", bConnectOnSIT ? "C" : "Do not c");

   if(localCallerID != NULL)
   {
      pInfo->ivLog->Print("      send our Caller ID as '%s'\n", localCallerID);
   }

   if((iSilenceTimeout == 0) && (iMaxVoiceTimeout == 0))
   {
      pInfo->ivLog->Print("      Do not perform initial voice length detection.\n");
   }
   else
   {
      pInfo->ivLog->Print("      Determine initial voice length with silence timeout of %d ms,\n", iSilenceTimeout);
      pInfo->ivLog->Print("      and overall voice length no greater than %d ms\n", iMaxVoiceTimeout);
   }

   // PlaceCall functionality available on licensed TeleFlow only
   if(pInfo->portLicense == UNLICENSED)
   {
      pInfo->ivLog->Print ("      THIS PORT IS UNLICENSED\n");
      delete[] localCallerID;
      return(IVR_NOTSUPPORTED);
   }

	status = pInfo->ivBoard->PlaceCall(localPhone.c_str(),
                                      localCallerID,
                                      iRingCnt,
                                      bSameLine ? BoardGeneric::existingLine : BoardGeneric::newLine,
                                      iFlashDur,
                                      bConnectOnCED,
                                      bConnectOnSIT,
                                      pVoiceLength,
                                      iSilenceTimeout,
                                      iMaxVoiceTimeout,
                                      iCallingNumPlan,
                                      iCallingNumType,
                                      iCalledNumPlan,
                                      iCalledNumType);

   delete[] localCallerID;

   if(!bSameLine)
   {
      // If we initiated the call, then set call direction;
      // result of PlaceCall does not matter since an out-bound call was attempted
      pInfo->variables->SetSystem("@SYS_CALL_DIRECTION", "OUT");
   }

	if(status == IVR_SUCCESS)
   {
      if((iSilenceTimeout > 0) && (!voiceLengthVariable.empty()))
      {
         char number[14];
         sprintf(number, "%d", voiceLength);

         if(!pInfo->variables->Set(voiceLengthVariable.c_str(), number))
         {
            pInfo->ivLog->Print("      ERROR: Failed to set Initial Voice Lenght variable '%s'\n", voiceLengthVariable.c_str());
            return(IVR_FATAL);
         }
      }

      // If ISDN, retrieve call reference from board
      if(stricmp(pInfo->pszProtocol, "isd0") == 0)
      {
         WORD callRef = pInfo->ivBoard->GetCallReference();
         if (callRef > 0)
         {
            char pszCallRef[24];
            sprintf(pszCallRef, "%u", callRef);
            pInfo->variables->SetSystem("@SYS_CALLREF", pszCallRef);
         }
         else
         {
            pInfo->variables->SetSystem("@SYS_CALLREF", "");
         }
      }
   }

   return(status);
}
// StepPlaceCall::Execute


bool StepPlaceCall::EvaluateNumberPlan(const char* pszPlan, BoardGeneric::ISDNNumberPlan &enumPlan)
{
   if((stricmp(pszPlan, "DEFAULT") == 0) || (stricmp(pszPlan, "UNKNOWN") == 0))
   {
      enumPlan = BoardGeneric::UNKNOWN_NUMBER_PLAN;
   }
   else if(stricmp(pszPlan, "ISDN") == 0)
   {
      enumPlan = BoardGeneric::ISDN_NUMBER_PLAN;
   }
   else if(stricmp(pszPlan, "TELEPHONY") == 0)
   {
      enumPlan = BoardGeneric::TELEPHONY_NUMBER_PLAN;
   }
   else if(stricmp(pszPlan, "PRIVATE") == 0)
   {
      enumPlan = BoardGeneric::PRIVATE_NUMBER_PLAN;
   }
   else
   {
      // invalid
      return(false);
   }
   // valid
   return(true);
}

bool StepPlaceCall::EvaluateNumberType(const char* pszType, BoardGeneric::ISDNNumberType &enumType)
{
   if((stricmp(pszType, "DEFAULT") == 0) || (stricmp(pszType, "UNKNOWN") == 0))
   {
      enumType = BoardGeneric::UNKNOWN_NUMBER_TYPE;
   }
   else if(stricmp(pszType, "INTERNATIONAL") == 0)
   {
      enumType = BoardGeneric::INTERNATIONAL_NUMBER_TYPE;
   }
   else if(stricmp(pszType, "NATIONAL") == 0)
   {
      enumType = BoardGeneric::NATIONAL_NUMBER_TYPE;
   }
   else if(stricmp(pszType, "LOCAL") == 0)
   {
      enumType = BoardGeneric::LOCAL_NUMBER_TYPE;
   }
   else
   {
      // invalid
      return(false);
   }
   // valid
   return(true);
}

