#include <stdio.h>                  // sprintf

#include "..\escutil\iv_str.h"      // FindCaseInsensitiveString
#include "InternetMessage.h"

InternetMessage::InternetMessage(void)
{
}

InternetMessage::~InternetMessage(void)
{
}

void InternetMessage::AddText(const char* text)
{
   if(text == NULL)
   {
      return;
   }

   this->text += (string) text;
}

void InternetMessage::ClearText(void)
{
   text.clear();
}

string::size_type InternetMessage::FindField(const char* fieldName)
{
   string::size_type headerEnd;
   string::size_type index;

   if(fieldName == NULL)
   {
      return(string::npos);
   }

   // If the header separator does not exist, consider the end of the string as the end of the header
   if((headerEnd = text.find("\r\n\r\n")) == string::npos)
   {
      headerEnd = text.length();
   }

   // Search for field name
   for(index = FindCaseInsensitiveString(text.substr(0, headerEnd).c_str(), fieldName);
      // field name found
      (index != string::npos)
      // field name not followed by field name separator
      && (text[index + strlen(fieldName)] != ':')
      // field name not followed by white space
      && (text[index + strlen(fieldName)] != ' ')
      // field name not at beginning of header or
      && (index != 0)
      // field name not preceded by field separator
      && ((index < 2) || (text[index - 1] != '\n') || (text[index - 2] != '\r'));
      index = FindCaseInsensitiveString(text.substr(index + 1, headerEnd).c_str(), fieldName));

   return(index);
}

const char* InternetMessage::GetBody(void)
{
   string::size_type bodyStart = text.find("\r\n\r\n");

   // If there is no header separator, there is no body, so return empty string
   if(bodyStart == string::npos)
   {
      return("");
   }
   else
   {
      bodyStart += 4;
   }

   body = text.substr(bodyStart);

   return(body.c_str());
}

const char* InternetMessage::GetField(const char* fieldName)
{
   string::size_type headerEnd;
   string::size_type index1;
   string::size_type index2;

   if(fieldName == NULL)
   {
      return(NULL);
   }

   // fieldName not found in the header
   if((index1 = FindField(fieldName)) == string::npos)
   {
      return(NULL);
   }

   // If the header separator does not exist, consider the end of the string as the end of the header
   if((headerEnd = text.find("\r\n\r\n")) == string::npos)
   {
      headerEnd = text.length();
   }

   // Go to field name separator
   if((index1 = text.find(':', index1 + strlen(fieldName) + 1)) == string::npos)
   {
      return("");
   }

   // Find start of field text
   for(index1++; (index1 < headerEnd) && (text[index1] == ' '); index1++);

   // Find field separator
   if((index2 = text.substr(index1, headerEnd).find("\r\n")) == string::npos)
   {
      index2 = headerEnd;
   }

   fieldText = text.substr(index1, index2);

   return(fieldText.c_str());
}

const char* InternetMessage::GetHeader(void)
{
   string::size_type headerEnd = text.find("\r\n\r\n");

   // Create header separator if missing
   if(headerEnd == string::npos)
   {

      text += (string) "\r\n\r\n";
      headerEnd = text.length();
   }
   else
   {
      headerEnd += 4;
   }

   // Include header separator
   header = text.substr(0, headerEnd);

   return(header.c_str());
}

const char* InternetMessage::GetText(void) const
{
   return(text.c_str());
}

void InternetMessage::RemoveField(const char* fieldName)
{
   string::size_type headerEnd;
   string::size_type fieldBegin;
   string::size_type fieldEnd;

   if(fieldName == NULL)
   {
      return;
   }

   // fieldName not found in the header
   if((fieldBegin = FindField(fieldName)) == string::npos)
   {
      return;
   }

   // If the header separator does not exist, consider the end of the string as the end of the header
   if((headerEnd = text.find("\r\n\r\n")) == string::npos)
   {
      headerEnd = text.length();
   }

   // Find field separator
   if((fieldEnd = text.substr(fieldBegin, headerEnd).find("\r\n")) == string::npos)
   {
      fieldEnd = headerEnd;
   }

   text.erase(fieldBegin, fieldEnd);
}

void InternetMessage::SetBody(const char* text)
{
   string::size_type bodyStart = this->text.find("\r\n\r\n");
   char number[14];

   // Create header separator at end of message if missing
   if(bodyStart == string::npos)
   {
      this->text += (string) "\r\n\r\n";
      bodyStart = this->text.find("\r\n\r\n") + 4;
   }
   else
   {
      bodyStart += 4;
   }

   if(text == NULL)
   {
      text = "";
   }

   // Replace current body
   this->text = this->text.substr(0, bodyStart) + (string) text;

   if(strlen(text) == 0)
   {
      RemoveField("Content-Length");
   }
   else
   {
      // Update content-length field
      sprintf(number, "%d", this->text.length() - bodyStart);
      SetField("Content-Length", number);
   }
}

void InternetMessage::SetField(const char* fieldName, const char* fieldText)
{
   string::size_type headerEnd;
   string::size_type index1;
   string::size_type index2;
   string newField;

   if(fieldName == NULL)
   {
      return;
   }

   // Create field entry
   newField = (string) fieldName + (string) ": " + (string) (fieldText == NULL ? "" : fieldText) + (string) "\r\n";

   // Field lines should never be more than 1000 characters including the field
   // separator, as per RFC 2822
   newField = newField.substr(0, 1000);

   // Truncate after first field separator to retain header integrity
   newField = newField.substr(0, newField.find("\r\n") + 2);

   // Create header separator if missing
   if((headerEnd = text.find("\r\n\r\n")) == string::npos)
   {
      text += (string) "\r\n\r\n";
      headerEnd = text.find("\r\n\r\n");
      text = text.substr(0, headerEnd + 4);
   }

   // fieldName does not exist
   if((index1 = FindField(fieldName)) == string::npos)
   {
      // There are no fields yet (header separator is right at the beginning)
      // Over-write first "\r\n" of the header separator
      if(headerEnd == 0)
      {
         text.replace(text.begin(), text.begin() + 2, newField);
      }
      // There are already some fields, insert new field at end
      else
      {
         text.insert(headerEnd + 2, newField);
      }
   }
   // fieldName already exists
   else
   {
      // Replace current field
      index2 = text.substr(index1).find("\r\n") + 2;
      text.replace(text.begin() + index1, text.begin() + index1 + index2, newField);
   }
}

void InternetMessage::SetHeader(const char* text)
{
   string newHeader;
   string::size_type bodyStart;

   if(text == NULL)
   {
      newHeader = "";
   }
   else
   {
      newHeader = text;
   }

   // Ensure that new header ends with proper header separator
   if((bodyStart = newHeader.find("\r\n\r\n")) != string::npos)
   {
      newHeader = newHeader.substr(0, bodyStart);
   }

   newHeader += "\r\n\r\n";

   // Find current header separator
   if((bodyStart = this->text.find("\r\n\r\n")) == string::npos)
   {
      bodyStart = this->text.length();
   }
   else
   {
      bodyStart += 4;
   }

   // Replace current header
   this->text.replace(this->text.begin(), this->text.begin() + bodyStart, newHeader);
}

void InternetMessage::SetText(const char* text)
{
   if(text == NULL)
   {
      this->text.clear();
   }
   else
   {
      this->text = text;
   }
}

