#ifndef __INTERNETMESSAGE_H__
#define __INTERNETMESSAGE_H__

#ifdef __DLL__
#define INTERNETMESSAGEAPI __declspec(dllexport)
#else
#define INTERNETMESSAGEAPI __declspec(dllimport)
#endif

#include <string>
using std::string;

// Class for managing Internet messages as per RFC 2822
class INTERNETMESSAGEAPI InternetMessage
{
   public:

      InternetMessage(void);
      virtual ~InternetMessage(void);

      // Appends text to end of the message
      // Performs no format validity checking
      // Has no effect iff text is NULL
      virtual void AddText(const char* text);

      // Clears all text from the message
      // GetBody, GetHeader, and GetText will now return an empty string
      // GetField will return NULL
      virtual void ClearText(void);

      // Returns the offset of fieldName in the header
      // Returns string::npos if fieldName is not in the header
      virtual string::size_type FindField(const char* fieldName);

      // Returns the body portion of the message
      // The const char* is only valid until the next GetBody call
      virtual const char* GetBody(void);

      // Returns the value of fieldName from the header
      // fieldName is case-insensitive
      // Returns NULL if fieldName is not in the header
      // Returns NULL if fieldName is NULL
      // The const char* is only valid until the next GetField call
      virtual const char* GetField(const char* fieldName);

      // Returns the header portion of the message
      // The const char* is only valid until the next GetHeader call
      virtual const char* GetHeader(void);

      // Returns the entire message
      // The const char* is only valid until the message text is changed
      virtual const char* GetText(void) const;

      // Removes fieldName from the header
      // Has no effect if fieldName is NULL, or fieldName does not exist in the header
      virtual void RemoveField(const char* fieldName);

      // Sets the body portion of the message to text
      // Creates a header separator at end of message first, if needed
      // Sets the body to an empty string if text is NULL
      virtual void SetBody(const char* text);

      // Sets the value of fieldName in the header
      // fieldName is case-insensitive
      // Has no effect if fieldName is NULL
      // Sets value to an empty string if fieldText is NULL
      // Does not check for a valid fieldName, other than NULL check
      // Does not check for valid fieldText, other than NULL check
      // If fieldName is not in the header, it is added to the end of the header
      // Creates a header separator at end of message first, if needed
      // Field lines ("<fieldName>: <fieldText>\r\n") must never be more than 1000
      // characters, as per RFC 2822, and will be truncated to 1000 characters
      virtual void SetField(const char* fieldName, const char* fieldText);

      // Sets the header portion of the message to text
      // Creates a header separator if not included in text
      // Performs no header format validity checking other than header separator
      // Sets the header to an empty string if text is NULL
      virtual void SetHeader(const char* text);

      // Sets the message to text
      // Performs no format validity checking
      // Clears the message if text is NULL
      virtual void SetText(const char* text);

   protected:

      string text;
      string body;
      string fieldText;
      string header;
};

#endif

